# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html
# credits: http://www.opsi.org/credits/

#***************************************************************************
# Subversion:
# $Revision$
# $Author$
# $Date$
#***************************************************************************

#

try:
    import ConfigParser
except:
    import configparser
from OPSI.Util.File import IniFile
from OPSI.Util.File import TextFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from OPSI.Util import blowfishDecrypt
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse
from os.path import splitext, basename, isfile
import platform
import re, xml.dom.minidom, os, gzip
import crypt
import codecs
from OpenSSL.crypto import FILETYPE_PEM, load_certificate
from base64 import b64encode
from os import urandom

if not bootimageVersion or (int(bootimageVersion) < 20210618):
    raise Exception(u"This product requires opsi-linux-bootimage version >= 20210618")
try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

opsiVersion = backend.backend_info()['opsiVersion']

# Do hardware inventory
logger.notice(u"Fetching opsi hw audit configuration")
hwconfig = backend.auditHardware_getConfig()
if isOpsi43:
    show_message(u"Running hardware inventory", 10)
logger.notice(u"Running hardware inventory")
auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

logger.notice(u"Sending hardware information to service")
backend.auditHardwareOnHost_setObsolete(clientId)
backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsipreparelib.py','/tmp')
exec(open('/tmp/opsipreparelib.py').read())

initOpsiSetupLib()

inUefi = inUefiMode()
useGpt = False
if inUefi:
    useGpt = True
    currentbootentry = uefiGetActBootEntry()
    if currentbootentry:
        uefiNetbootLabel = currentbootentry['label']
        logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
        backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label")
        backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
    else:
        logger.notice('No usable current (net) boot entry found')

# create network path
if (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
    match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
    if not match:
        raise Exception("Bad depot-URL '%s'" % depot._url)
    hn = match.group(1)
    sn = match.group(2)
    pn = match.group(3)

depotDir = '/var/lib/opsi/depot'

swapSizeProp = productProperties.get('swap_size', '')
if swapSizeProp:
    removedUnits = removeUnit(swapSizeProp)
    swapSize = int(removedUnits)/1048576
    logger.notice(u"swap size: %s" % swapSize)
else:
    meminfo = open('/proc/meminfo').read()
    matched = re.search(r'^MemTotal:\s+(\d+)', meminfo)
    if matched:
        memory = int(matched.groups()[0])
    swapSize = int(memory)/1024
    logger.notice(u"memory: %s" % swapSize)

proxy = productProperties.get('proxy','')

install_opsi_client_agent = False
do_opsi_client = "false"
if (str(productProperties.get('install_opsi-client-agent', 'true')).lower() == 'true'):
    install_opsi_client_agent = True
    do_opsi_client = "true"

install_unattended = True
if (str(productProperties.get('install_unattended', 'true')).lower() == 'false'):
    install_unattended = False

install_desktop = True
if (str(productProperties.get('install_desktop', 'true')).lower() == 'false'):
    install_desktop = False

language = productProperties.get('language','de_DE')
keymap = productProperties.get('console_keymap', 'german')

my_broadcast = getBroadcastAddress(ipaddress, netmask)
my_network = getNetworkAddress(ipaddress, netmask)

suse_register = False
do_suse_registration = 'false'
if (str(productProperties.get('suse_register', 'true')).lower() == 'true'):
    suse_register = True
logger.notice(u"suse_register:%s" % suse_register)
if suse_register:
    logger.notice(u"softwareLicenseKey:%s" % softwareLicenseKey)
    if (softwareLicenseKey == ''):
        raise Exception("No Licensekey found")
    else:
        do_suse_registration = 'true'
        logger.notice(u"splitting softwareLicenseKey for suse_register")
        registration_email=softwareLicenseKey.split(':')[0]
        registration_key=softwareLicenseKey.split(':')[1]
        registration_key=registration_key.upper()

# Get and encrypt passwords
# user
userpassclear = productProperties.get('user_password', 'linux123')
userpassenc = shadow_password_hash(userpassclear)

# root
rootpassclear = productProperties.get('root_password', 'linux123')
rootpassenc = shadow_password_hash(rootpassclear)

if (proxy == ''):
    do_proxy = 'true'
    proxy_addr = proxy
else:
    do_proxy = 'false'
    proxy_addr = ''

# get machine architecture
myarchitecture = getArchitecture()
if myarchitecture == 'amd64':
    scriptMessageSubject.setMessage(u"Installing 64 Bit system....")
    logger.notice(u"Installing 64 Bit system....")
    myarchitectureBits = '64'
else:
    scriptMessageSubject.setMessage("Installing 32 Bit system....")
    logger.notice(u"Installing 32 Bit system....")
    myarchitectureBits = '32'

if isOpsi43:
    show_message("Preparing disk", 20)
scriptMessageSubject.setMessage("Create temporary place on disk.... ")
logger.notice(u"Create temporary place on disk.... ")
#prepareDisk()
target = '/mnt/hd'
systemPartitionNumber = 1
systemPartitionSizeM = 2000
systemPartitionLabel  = 'LINUX'
blockAlignment = True
useGpt = False

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
execute('sfdisk -l')

# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# use first disk
disk = disks[0]
logger.notice(u"using sgdisk to clear partition table of %s" % (disk.device))
execute('/usr/sbin/sgdisk --zap-all %s' % (disk.device))
fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
disk.deletePartitionTable()
fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ext4", gptFileSystem="8300")
createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ext4", disk.device, useGpt)
systemPartitionDevicePath = getPartitionDevicePath(disk.device, systemPartitionNumber)

# Partition table deleted => delete all netboot product states
deletePocs = []
for poc in backend.productOnClient_getObjects(clientId = clientId):
    if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
       ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
        deletePocs.append(poc)
backend.productOnClient_deleteObjects(deletePocs)

diskSizeM = (int(disk.size)/(1024*1024))
if inUefi:
    rootSize = diskSizeM-swapSize-150
else:
    rootSize = diskSizeM-swapSize

diskSizeM = str(diskSizeM) + ' MB'
rootSize = str(rootSize) + ' MB'
swapSize = str(swapSize) + ' MB'

logger.notice(u"disk: %s, disk.size: %s, root.size: %s, swap.size: %s" %(disk.device, diskSizeM, rootSize, swapSize))

if isOpsi43:
    show_message("Mounting partitions and copying data", 40)
# Mount temporary partition
checkCreateDir(target)
checkClearMountpoint(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkClearPartition(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkMount(target, disk.device, systemPartitionNumber, "ext4", systemPartitionDevicePath)
checkCreateDir(target+'/tmp')

scriptMessageSubject.setMessage("Getting kernel and initrd.... ")
logger.notice(u"Getting kernel and initrd.... ")
try:
    depot.copy(SCRIPT_PATH + '/isocontent/boot/x86_64/loader/linux', target+'/tmp')
    depot.copy(SCRIPT_PATH + '/isocontent/boot/x86_64/loader/initrd', target+'/tmp')
except Exception as fnferror:
    raise Exception("isocontent directory doesn't seem to be filled")

scriptMessageSubject.setMessage("Uncompressing initrd.... ")
logger.notice(u"Uncompressing initrd.... ")
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("mv %s/tmp/initrd %s/tmp/initrd.xz" % (target,target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("xz -d %s/tmp/initrd.xz"% (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

scriptMessageSubject.setMessage("Patching initrd.... ")
logger.notice(u"Mounting initrd ....")
execute("mkdir -p %s/tmp/mnt" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd" % (target))

logger.notice(u"Copy profile for patch....")
depot.copy(SCRIPT_PATH+'/autoyast.xml', target+'/tmp/installation_profile')
if os.path.exists(DEPOT_MOUNT + '/' + SCRIPT_PATH+'/custom/autoyast.xml'):
    depot.copy(SCRIPT_PATH+'/custom/autoyast.xml', target+'/tmp/installation_profile')

if install_desktop:
    default_target = 'graphical'
else:
    default_target = 'multi-user'
    logger.notice("Remove X11 and Desktop from xml file ....")
    profile = open(target+'/tmp/installation_profile',"r+")
    profilelines = profile.readlines()
    profile.seek(0)
    for aktline in profilelines:
        if (aktline.strip() != "<pattern>gnome-basic</pattern>") and (aktline.strip() != "<pattern>x11</pattern>"):
            profile.write(aktline)
    profile.truncate()
    profile.close()

depotAddress = depotUrl.split('/')[2].replace(':4447','')
execute(f"wget --no-check-certificate -P /tmp/ https://{fqdn}:{pckey}@{depotAddress}:4447/ssl/opsi-ca-cert.pem")
execute(f"cat /tmp/opsi-ca-cert.pem >> {target}/tmp/mnt/var/lib/ca-certificates/ca-bundle.pem")
execute(f"cp /tmp/opsi-ca-cert.pem {target}/tmp/mnt/var/lib/ca-certificates/pem/")
cert_file = "/tmp/opsi-ca-cert.pem"
with codecs.open(cert_file, "r", "utf-8") as file:
    c = load_certificate(FILETYPE_PEM, file.read().encode("ascii"))
    hex_hash = f"{c.subject_name_hash():x}"
if hex_hash:
    execute(f"cd {target}/tmp/mnt/var/lib/ca-certificates/pem/ && ln -s opsi-ca-cert.pem {hex_hash}.0")
    execute(f"ls -hal {target}/tmp/mnt/var/lib/ca-certificates/pem/{hex_hash}.0")

# Copy opsi-client-agent
depot.copy('/opsi-linux-client-agent/files/opsi/cfg/config.ini', target + '/tmp/')
#
logger.notice(u"Copy SYSCONF_INI....")
# Copy sysconf.ini to opsi-client-agent config dir
copy(SYSCONF_INI , target +'/tmp/')

logger.notice(u"Prepare patch input....")
with open(PATCHA_IN, "a") as myfile:
    myfile.write('networkaddress=%s\n' % my_network)
    myfile.write('broadcastaddress=%s\n' % my_broadcast)
    myfile.write('nameserveraddress=%s\n' % nameserver)
    myfile.write('do_suse_registration=%s\n' % do_suse_registration)
    myfile.write('registration_email=%s\n' % registration_email)
    myfile.write('registration_key=%s\n' % registration_key)
    myfile.write('do_opsi_client=%s\n' % do_opsi_client)
    myfile.write('userpassenc=%s\n' % userpassenc)
    myfile.write('rootpassenc=%s\n' % rootpassenc)
    myfile.write('do_proxy=%s\n' % do_proxy)
    myfile.write('proxy_addr=%s\n' % proxy_addr)
    myfile.write('default_target=%s\n' % default_target)
    myfile.write('disk=%s\n' % disk.device)
    myfile.write('rootsize=%s\n' % rootSize)
    myfile.write('swapsize=%s\n' % swapSize)
    myfile.write('service_ip_address=%s\n' % (nextserver.replace("/rpc", "").rstrip("/")))
    myfile.write('depotAddress=%s\n' % (depotUrl.split('/')[2].replace(':4447','')))
    myfile.write('client_id=%s\n' % (fqdn))
    myfile.write('client_key=%s\n' % (pckey))
    myfile.write('depotAddress=%s\n' % (depotUrl.split('/')[2].replace(':4447','')))

logger.notice(u"Here comes the PATCHA_IN (%s):" % PATCHA_IN)
execute('cat %s' % (PATCHA_IN))

if isOpsi43:
    show_message("patching files", 90)
logger.notice(u"Patching installation_profile")
if install_opsi_client_agent == False:
    lines = []
    with open(target + '/tmp/installation_profile', 'r') as inputxml:
        lines = inputxml.readlines()
    outlines = []
    keep = True
    for line in lines:
        if "mkdir -p /var/tmp/opsi-linux-client-agent" in line:
            keep = False
        if keep:
            outlines.append(line)
        if "systemctl enable olca-install.service" in line:
            keep = True
    with open(target + '/tmp/installation_profile', 'w') as outputxml:
        outputxml.writelines(outlines)

logger.notice(u"Patching installation_profile and config.ini")
for patchfile in ( target + '/tmp/config.ini', target + '/tmp/installation_profile' ):
    execute('/usr/local/bin/patcha %s' % patchfile)
    execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, patchfile))

logger.notice(u"Copy opsinetmount file to initrd...")
depot.copy(SCRIPT_PATH+'/opsinetmount', target+'/tmp/mnt/sbin')
execute("chmod 777 %s/tmp/mnt/sbin/opsinetmount" % (target))

if install_unattended:
    logger.notice(u"Copy profile file to initrd...")
    copy(target+'/tmp/installation_profile', target+'/tmp/mnt/autoinst.xml')
    logger.notice(u"Here comes the installation profile:")
    execute('cat %s/tmp/mnt/autoinst.xml' % (target))

scriptMessageSubject.setMessage("Rebuild patched initrd.... ")
logger.notice(u"Rebuild patched initrd.... ")
execute("rm %s/tmp/initrd" % (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("cd %s/tmp/mnt ; find . | cpio -o -H newc > %s/tmp/initrd" % (target, target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

# we do not compress initrd again
#execute("xz %s/tmp/initrd" % (target))
#try:
#       execute("ls -l %s/tmp/initrd*" % (target))
#except:
#       pass
#execute("mv %s/tmp/initrd.xz %s/tmp/initrd" % (target,target))
#try:
#       execute("ls -l %s/tmp/initrd*" % (target))
#except:
#       pass


scriptMessageSubject.setMessage("Prepare loading kernel.... ")
logger.notice(u"Prepare loading kernel ....")
kernel_file = target+'/tmp/linux'
append_line = "splash=silent "
inird_file = target+'/tmp/initrd'
#depot.copy(SCRIPT_PATH+'/kexec','/sbin')
#execute("chmod u+x /sbin/kexec")
execute(f"{which('kexec')} --load {kernel_file}  --append='install=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/isocontent/ {append_line}' --initrd={inird_file}")

logger.notice(u"handle_setupAfterInstall ....")
handle_setupAfterInstall()

# Set product installation status
productOnClient.setInstallationStatus('unknown')
#productOnClient.setActionProgress('installing')
productOnClient.setActionResult('')
productOnClient.setLastAction('setup')

#scriptMessageSubject.setMessage("Writing log to service.... ")
try:
    # Write log to service
    logger.notice(u"Writing log to service")

    log = codecs.open(LOG_FILE, 'r', 'utf-8', 'replace')
    data = log.read()
    log.close()

    logger.setFileLevel(LOG_ERROR)
    if backend:
        try:
            try:
                ui.getMessageBox().addText(_(u"Writing log to service\n"))
            except:
                pass
            backend.log_write(logType = 'bootimage', data = data, objectId = clientId, append = False)
        except Exception as e:
            logger.error(u"Failed to write log to service: %s" % e)

except Exception as e:
    logger.logException(e)

scriptMessageSubject.setMessage("Execute new kernel.... ")
logger.notice(u"Execute new kernel ....")
execute("%s -e" % (which('kexec')))
#
